"""
#===============================================================================
#
#  License: GPL
#
#
#  Copyright (c) 2019 Rob Serafin, Liu Lab,
#  The University of Washington Department of Mechanical Engineering

#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#===============================================================================

Rob Serafin
3/25/2020

"""

import numpy
from numba import njit
import matplotlib.pyplot as plt


@njit
def sortImage(image, mask_val=255, greater_mode=False):
    """
    Method for sorting image pixel values excluding a high value
    threshold. Used by getRGBStats to get a 1D array for pixel values
    less than saturated value.

    Parameters
    ----------

    image : 2D numpy array
        Image array

    mask_val : int
        High threshold over which pixel values will be ignored

    greater_mode: bool
        default = False, whether to filter out values less than mask_val


    Returns
    -------
    pixel_set : 1D numpy array
        Pixel values contained in image which are under the mask value.
    """
    pixel_set = []

    # Ravel image and sort through item by item
    image_list = image.ravel()
    for item in image_list:
        if not greater_mode:
            if item < mask_val:
                # append to pixel set if value is under threshold
                pixel_set.append(item)
        elif greater_mode:
            if item > mask_val:
                pixel_set.append(item)
        else:
            if item != mask_val:
                pixel_set.append(item)
    return numpy.asarray(pixel_set)


def getRGBStats(image, mask_val=255):
    """
    Method which returns dictionary containing useful image statistics
    from the input RGB image. Will be called from a save metadata method
    which will store data in a json.


    Parameters
    ----------

    image : 3D numpy array
        RGB image array

    mask_val : int
        High threshold over which pixel values will be ignored

    Returns
    -------

    image_stats : dict
        Dictionary with 'R', 'G', 'B' keys which each have a dict as
        follows:

            'data' : data called from sortImage method, below mask value

            'median' : median of sorted image data

            '90th' : 90th percentile of sorted image data

            '10th' : 10th percentile of sorted image data

    """
    R = {'data': sortImage(image[:, :, 0], mask_val=mask_val)}
    G = {'data': sortImage(image[:, :, 1], mask_val=mask_val)}
    B = {'data': sortImage(image[:, :, 2], mask_val=mask_val)}

    R['median'] = numpy.median(R['data'])
    G['median'] = numpy.median(G['data'])
    B['median'] = numpy.median(B['data'])

    R['90th'] = numpy.percentile(R['data'], 90)
    R['10th'] = numpy.percentile(R['data'], 10)

    G['90th'] = numpy.percentile(G['data'], 90)
    G['10th'] = numpy.percentile(G['data'], 10)

    B['90th'] = numpy.percentile(B['data'], 90)
    B['10th'] = numpy.percentile(B['data'], 10)

    image_stats = {'R': R, 'G': G, 'B': B}

    return image_stats


def getHSVstats(nuclei, cyto,
                hue_mask_value=0,
                sat_mask_value=0,
                val_mask_value=0):
    """
    Method which returns dictionary containing useful image statistics
    from the input RGB image. Will be called from a save metadata method
    which will store data in a json.


    Parameters
    ----------

    nuclei : 3D numpy array
        Image array to analyze which has been converted from RGB to HSV
        space, representing the hematoxylin, or equivalent, segmented
        from a histology or virually stained image.

    cyto : 3D numpy array
        Image array to analyze which has been converted from RGB to HSV
        space, representing the eosin, or equivalent, segmented from a
        histology or virually stained image.

    mask_val : int
        High threshold over which pixel values will be ignored

    Returns
    -------

    image_stats : dict
        Dictionary with 'Hue', 'Sat', 'Val' keys which each have a dict
        as follows:

            'median' : median of sorted image data

            '90th' : 90th percentile of sorted image data

            '10th' : 10th percentile of sorted image data

            'std' : standard deviation of sorted image data
    """

    H_nuc = sortImage(nuclei[:, :, 0], mask_val=hue_mask_value,
                      greater_mode=True)

    S_nuc = sortImage(nuclei[:, :, 1], mask_val=sat_mask_value,
                      greater_mode=True)

    V_nuc = sortImage(nuclei[:, :, 2], mask_val=val_mask_value,
                      greater_mode=True)

    H_cyto = sortImage(cyto[:, :, 0], mask_val=hue_mask_value,
                       greater_mode=True)

    S_cyto = sortImage(cyto[:, :, 1], mask_val=sat_mask_value,
                       greater_mode=True)

    V_cyto = sortImage(cyto[:, :, 2], mask_val=val_mask_value,
                       greater_mode=True)

    image_stats = {'nuclei': {'Hue': {}, 'Sat': {}, 'Val': {}},
                   'cyto': {'Hue': {}, 'Sat': {}, 'Val': {}}}

    image_stats['nuclei']['Hue']['median'] = numpy.median(H_nuc)

    image_stats['nuclei']['Hue']['10'] = numpy.percentile(H_nuc, 10)

    image_stats['nuclei']['Hue']['90'] = numpy.percentile(H_nuc, 90)

    image_stats['nuclei']['Hue']['std'] = numpy.std(H_nuc)

    image_stats['nuclei']['Sat']['median'] = numpy.median(S_nuc)

    image_stats['nuclei']['Sat']['10'] = numpy.percentile(S_nuc, 10)

    image_stats['nuclei']['Sat']['90'] = numpy.percentile(S_nuc, 90)

    image_stats['nuclei']['Sat']['std'] = numpy.std(S_nuc)

    image_stats['nuclei']['Val']['median'] = numpy.median(V_nuc)

    image_stats['nuclei']['Val']['10'] = numpy.percentile(V_nuc, 10)

    image_stats['nuclei']['Val']['90'] = numpy.percentile(V_nuc, 90)

    image_stats['nuclei']['Val']['std'] = numpy.std(V_nuc)

    image_stats['cyto']['Hue']['median'] = numpy.median(H_cyto)

    image_stats['cyto']['Hue']['10'] = numpy.percentile(H_cyto, 10)

    image_stats['cyto']['Hue']['90'] = numpy.percentile(H_cyto, 90)

    image_stats['cyto']['Hue']['std'] = numpy.std(H_cyto)

    image_stats['cyto']['Sat']['median'] = numpy.median(S_cyto)

    image_stats['cyto']['Sat']['10'] = numpy.percentile(S_cyto, 10)

    image_stats['cyto']['Sat']['90'] = numpy.percentile(S_cyto, 90)

    image_stats['cyto']['Sat']['std'] = numpy.std(S_cyto)

    image_stats['cyto']['Val']['median'] = numpy.median(V_cyto)

    image_stats['cyto']['Val']['10'] = numpy.percentile(V_cyto, 10)

    image_stats['cyto']['Val']['90'] = numpy.percentile(V_cyto, 90)

    image_stats['cyto']['Val']['std'] = numpy.std(V_cyto)

    return image_stats


def ViewImage(Image, title=None, do_hist=False,
              figsize=(6, 4), range_min=0, range_max=None,
              cmap='viridis', do_ticks=False):

    """
    Method for viewing images easily.

    Parameters
    ----------

    Image : 2D numpy array
        Image to view.

    title : None or str
        default to None, if not will be used as plot title.

    do_hist : bool
        default to False, if True will plot histogram of image in an
        adjacent subplot.

    figsize : tuple
        defaults to (6,4), tuple of ints which defines size of plotted
        image.

    range_min : int
        defaults to 0, minimum range for histogram.

    range_max : None or int
        defaults to None, if int maximum range for histogram.

    cmap : str
        defaults to 'viridis', colormap for plotting grayscale images.

    do_ticks : bool
        defaults to False, plot x/y ticks and ticklabels on plot.

    Returns
    -------
    f,ax : matplotlib figure, matplotlib axis
        figure and axis

    """

    if do_hist:

        f, ax = plt.subplots(ncols=2, figsize=figsize)

        # plot image
        ax[0].imshow(Image, cmap=cmap, vmin=Image.min(),
                     vmax=Image.max())
        ax[0].set_title('Image')

        # setup histogram
        if range_max is None:
            range_max = Image.max()
        ax[1].hist(Image[Image != 0].ravel(), 256, [range_min, range_max])

        ax[1].set_title('Histogram')
        ax[1].set_xlabel('Intenstity, A.U')

        asp = numpy.diff(ax[1].get_xlim())[0] / numpy.diff(ax[1].get_ylim())[0]

        ax[1].set_aspect(asp)
        plt.subplots_adjust(wspace=0.5)

        # set title
        if title is not None:
            f.suptitle(title)

    else:

        # plot image
        f, ax = plt.subplots(figsize=figsize)
        ax.imshow(Image, cmap=cmap)

        # set title
        if title is not None:
            ax.set_title(title)

    if do_ticks is False:

        # adjust tick parameters
        plt.tick_params(axis='both', which='both', top=False,
                        bottom=False, left=False, right=False,
                        labelbottom=False, labelleft=False)
    return f, ax



"""
#===============================================================================
#
#  License: GPL
#
#
#  Copyright (c) 2019 Rob Serafin, Liu Lab,
#  The University of Washington Department of Mechanical Engineering
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#===============================================================================

Rob Serafin
3/25/2020

"""

import scipy.ndimage as nd
import skimage.filters as filt
import skimage.morphology as morph
from skimage.color import rgb2hed, rgb2hsv
import cv2
import numpy
from numba import cuda
import math


@cuda.jit  # direct GPU compiling
def rapidGetRGBframe(nuclei, cyto, output,
                      nuc_settings, cyto_settings,
                      k_nuclei, k_cyto):
    # TODO: implement array base normalization
    """
    GPU based exponential false coloring operation. Used by
    rapidFalseColor().

    Parameters
    ----------
    nuclei : 2D numpy array
        dtype = float
        Nuclear channel image, already pre processed

    cyto : 2d numpy array
        dtype = float
        Cytoplasm channel image, already pre processed.

    nuc_settings : float
        RGB constant for nuclear channel

    cyto_settings : float
        RGB constant for cyto channel

    k_nuclei : float
        Additional multiplicative constant for nuclear channel.
        Eventually will get removed once flat fielding is in place for
        all pseudo coloring methods.

    k_cyto: float
        Additional multiplicative constant for cytoplasmic channel.
        Eventually will get removed once flat fielding is in place for
        all pseudo coloring methods.
    """
    row, col = cuda.grid(2)

    # iterate through image and assign pixel values
    if row < output.shape[0] and col < output.shape[1]:
        tmp = nuclei[row, col]*nuc_settings*k_nuclei + \
                                            cyto[row, col]*cyto_settings*k_cyto
        output[row, col] = 255*math.exp(-1*tmp)


@cuda.jit
def rapidFieldDivision(image, flat_field, output):
    """
    Used for rapidFalseColoring() when flat field has been calculated

    Parameters
    ----------

    image : numpy array written to GPU

    flat_field : numpy array written to GPU

    output : numpy array written to GPU
        result from computation

    """
    row, col = cuda.grid(2)

    if row < output.shape[0] and col < output.shape[1]:
        tmp = image[row, col]/flat_field[row, col]
        output[row, col] = tmp


def rapidFalseColor(nuclei, cyto, nuc_settings, cyto_settings,
                    TPB=(32, 32),
                    nuc_normfactor=8500,
                    cyto_normfactor=2000,
                    run_FlatField_nuc=False,
                    run_FlatField_cyto=False,
                    nuc_bg_threshold=50,
                    cyto_bg_threshold=50):
    """
    Parameters
    ----------

    nuclei : numpy array
        Nuclear channel image.

    cyto : numpy array
        Cytoplasm channel image.

    nuc_settings : list
        Settings of RGB constants for nuclear channel. Should be in
        order R, G, B.

    cyto_settings : list
        Settings of RGB constants for cytoplasm channel. Should be in
        order R, G, B.

    nuc_normfactor : int or array
        Defaults to empirically determined constant to reduce
        saturation. Otherwise it should be a numpy array representing
        the true flat field image.

    cyto_normfactor : int or array
        Defaults to empirically determined constant to reduce
        saturation. Otherwise it should be a numpy array representing
        the true flat field image.

    TPB : tuple (int,int)
        THREADS PER BLOCK: (x_threads,y_threads) used for GPU threads.

    run_FlatField : bool
        defaults to False, boolean to apply flatfield

    nuc_bg_threshold = int
        defaults to 50, threshold level for calculating nuclear
        background.

    cyto_bg_threshold = int
        defaults to 50, threshold level for calculating cytoplasmic
        background.



    Returns
    -------
    RGB_image : 3D numpy array
        Combined false colored image in the standard RGB format
        [X, Y, C].

    """

    # ensure float dtype
    nuclei = numpy.ascontiguousarray(nuclei, dtype=float)
    cyto = numpy.ascontiguousarray(cyto, dtype=float)

    # set mulciplicative constants
    k_nuclei = 1.0
    k_cyto = 1.0

    # create blockgrid for gpu
    blockspergrid_x = int(math.ceil(nuclei.shape[0] / TPB[0]))
    blockspergrid_y = int(math.ceil(nuclei.shape[1] / TPB[1]))
    blockspergrid = (blockspergrid_x, blockspergrid_y)

    # allocate memory for background subtraction
    pre_nuc_output = cuda.device_array(nuclei.shape)
    nuc_global_mem = cuda.to_device(nuclei)

    pre_cyto_output = cuda.device_array(cyto.shape)
    cyto_global_mem = cuda.to_device(cyto)

    # run background subtraction or normalization for nuclei

    # use intensity leveling
    if run_FlatField_nuc:
        nuc_normfactor = numpy.ascontiguousarray(nuc_normfactor)
        nuc_norm_mem = cuda.to_device(nuc_normfactor)

        rapidFieldDivision[blockspergrid, TPB](nuc_global_mem, nuc_norm_mem,
                                               pre_nuc_output)

    # otherwise use standard background subtraction
    else:
        k_nuclei = 0.08
        nuc_background = getBackgroundLevels(nuclei,
                                             threshold=nuc_bg_threshold)[1]

        rapidPreProcess[blockspergrid, TPB](nuc_global_mem, nuc_background,
                                             nuc_normfactor, pre_nuc_output)

    # run background subtraction or normalization for cyto

    # use intensity leveling
    if run_FlatField_cyto:
        cyto_normfactor = numpy.ascontiguousarray(cyto_normfactor)
        cyto_norm_mem = cuda.to_device(cyto_normfactor)
        rapidFieldDivision[blockspergrid, TPB](cyto_global_mem,
                                               cyto_norm_mem, pre_cyto_output)

    # otherwise use standard background subtraction
    else:
        k_cyto = 0.012
        cyto_background = getBackgroundLevels(cyto,
                                              threshold=cyto_bg_threshold)[1]

        rapidPreProcess[blockspergrid, TPB](cyto_global_mem,
                                             cyto_background,
                                             cyto_normfactor,
                                             pre_cyto_output)

    # create output array to iterate through
    output_global = cuda.device_array((3, nuclei.shape[0], nuclei.shape[1]),
                                      dtype=numpy.uint8)

    # iterate through output and assign values based on RGB settings
    for i, z in enumerate(output_global):

        # get color frame
        rapidGetRGBframe[blockspergrid, TPB](pre_nuc_output,
                                             pre_cyto_output,
                                             z,
                                             nuc_settings[i],
                                             cyto_settings[i],
                                             k_nuclei,
                                             k_cyto)

    RGB_image = output_global.copy_to_host()

    # reorder array to dimmensional form [X,Y,C]
    RGB_image = numpy.moveaxis(RGB_image, 0, -1)

    return RGB_image.astype(numpy.uint8)


@cuda.jit  # direct GPU compiling
def rapidPreProcess(image, background, norm_factor, output):
    """
    Background subtraction optimized for GPU, used by rapidFalseColor.

    Parameters
    ----------

    image : 2d numpy array, dtype = int16
        Image for background subtraction.

    background : int
        Constant for subtraction.

    norm_factor : int
        Empirically determaned constant for normalization after
        subtraction. Helps prevent saturation.

    output : 2d numpy array
        Numpy array of zeros for GPU to assign values to.

    Returns
    -------
    This method requires an output array as an argument,
    the results of the compuation are stored there.
    """

    # create iterator for gpu
    row, col = cuda.grid(2)

    # cycle through image shape and assign values
    if row < output.shape[0] and col < output.shape[1]:

        # subtract background and raise to factor
        tmp = image[row, col] - background

        # remove negative values
        if tmp < 0:
            output[row, col] = 0

        # normalize to 8bit range
        else:
            tmp = (tmp**0.85)*(65535/norm_factor)*(255/65535)
            output[row, col] = tmp


def falseColor(nuclei, cyto,
               output_dtype=numpy.uint8,
               nuc_bg_threshold=50,
               cyto_bg_threshold=50,
               nuc_normfactor=5000,
               cyto_normfactor=2000,
               color_key='HE',
               color_settings=None):
    """
    CPU-based two channel virtual H&E coloring using Beer's law method.


    Parameters
    ----------
    nuclei : 2D numpy array
       Image of nuclear stain (hematoxylin equivalent) for false
       coloring.

    cyto : 2D numpy array
        Image of cytoplasm stain (eosin equivalent) or antibody for
        false coloring.

    output_dtype : numpy.uint8
        output datatype for final RGB image

    nuc_bg_threshold : int
        defaults to 50, threshold level for calculating nuclear
        background.

    cyto_bg_threshold : int
        defaults to 50, threshold level for calculating cytoplasmic
        background.

    nuc_normfactor : None or int
        defaults to None, color saturation level for nuclear channel
        used in preProcess. Larger values will result in colors which
        are less vibrant.

    cyto_normfactor : None or int
        defaults to None, color saturation level for cytoplasmic or
        antibody channel used in preProcess. Larger values will result
        in colors which are less vibrant.


    color_key : str
        defaults to HE, color settings key for getColorSettings

    color_settings : None or dict
        defaults to None. Color settings for false coloring, if None
        color settings will be assigned by getColorSettings() with the
        color_key provided. If different color settings are desired the
        keys to the dictionary should be 'nuclei' and 'cyto'.


    Returns
    -------
    RGB_image : numpy array
        Combined virtual H&E image in the standard RGB format [X, Y, C]

    """
    beta_dict = {
                # adjustment for nuclear channel
                'K_nuclei': 0.08,

                # adjustment for cytoplasmic channel
                'K_cyto': 0.0120}

    # entries are lists in order of RGB constants
    if color_settings is None:
        color_settings = getColorSettings(key=color_key)

    constants_nuclei = color_settings['nuclei']
    k_nuclei = beta_dict['K_nuclei']

    constants_cyto = color_settings['cyto']
    k_cytoplasm = beta_dict['K_cyto']

    # execute background subtraction
    nuclei = nuclei.astype(float)
    nuclei = preProcess(nuclei, threshold=nuc_threshold,
                        normfactor=nuc_normfactor)

    cyto = cyto.astype(float)
    cyto = preProcess(cyto, threshold=cyto_threshold,
                      normfactor=cyto_normfactor)

    # create array to store RGB image
    RGB_image = numpy.zeros((3, nuclei.shape[0], nuclei.shape[1]))

    # iterate throough RGB constants and execute image multiplication
    for i in range(len(RGB_image)):
        tmp_c = constants_cyto[i]*k_cytoplasm*cyto
        tmp_n = constants_nuclei[i]*k_nuclei*nuclei
        RGB_image[i] = 255*numpy.multiply(numpy.exp(-tmp_c), numpy.exp(-tmp_n))

    # reshape to [X,Y,C]
    RGB_image = numpy.moveaxis(RGB_image, 0, -1)

    # rescale to 8bit range
    return RGB_image.astype(output_dtype)


def preProcess(image, threshold=50, normfactor=None):
    """
    Method used for background subtracting data with a fixed value

    Parameters
    ----------

    image : 2D numpy array
        image for processing

    threshold : int
        background level to subtract

    Returns
    -------

    processed_image : 2D numpy array
        Background subtracted image.
    """

    # background subtraction
    image -= threshold

    # no negative values
    image[image < 0] = 0

    # calculate normalization factor
    image = numpy.power(image, 0.85)
    if normfactor is None:
        normfactor = numpy.mean(image[image > threshold])*8

    # convert into 8bit range
    processed_image = image*(65535/normfactor)*(255/65535)

    return processed_image


@cuda.jit
def Convolve2d(image, kernel, output):
    """
    GPU based 2d convolution method.

    Parameters
    ----------

    image : 2D numpy array
        Image for processing, written to GPU.

    kernel : 2D numpy array
        kernel to convolve image with, written to GPU

    output : 2D numpy array
        Output array, storing result of convolution, written to GPU.

    Returns
    -------
    This method requires an output array as an argument, the results
    of the compuation are stored there.
    """

    # create iterator
    row, col = cuda.grid(2)

    image_rows, image_cols = image.shape

    delta_r = kernel.shape[0]//2
    delta_c = kernel.shape[1]//2

    # ignore rows/cols outside image
    if (row >= image_rows) or (col >= image_cols):
        return

    tmp = 0
    for i in range(kernel.shape[0]):
        for j in range(kernel.shape[1]):
            # result should be sum of kernel*image as kernel is varied
            row_i = row - i + delta_r
            col_j = col - j + delta_c
            if (row_i >= 0) and (row_i < image_rows):
                if (col_j >= 0) and (col_j < image_cols):
                    tmp += kernel[i, j]*image[row_i, col_j]

    output[row, col] = tmp


def sharpenImage(input_image, alpha=0.5):
    """
    Image sharpening algorithm to amplify edges.

    Parameters
    ----------

    input_image : 2D numpy array
        Image to run sharpening algorithm on

    alpha : float or int
        Multiplicative constant for final result.

    Returns
    --------

    final_image : 2D numpy array
        The sum of the input image and the resulting convolutions
    """
    # create kernels to amplify edges
    hkernel = numpy.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]])
    vkernel = numpy.array([[1, 0, -1], [1, 0, -1], [1, 0, -1]])

    # set grid/threads for GPU
    blocks = (32, 32)
    grid = (input_image.shape[0]//blocks[0] + 1,
            input_image.shape[1]//blocks[1] + 1)

    # run convolution
    input_image = numpy.ascontiguousarray(input_image)
    voutput = cuda.device_array(input_image.shape, dtype=numpy.float64)
    houtput = cuda.device_array(input_image.shape, dtype=numpy.float64)
    Convolve2d[grid, blocks](input_image, vkernel, voutput)
    Convolve2d[grid, blocks](input_image, hkernel, houtput)

    voutput = voutput.copy_to_host()
    houtput = houtput.copy_to_host()

    # calculate final result
    final_image = input_image + alpha*numpy.sqrt(voutput**2 + houtput**2)

    return final_image


def getColorSettings(key='HE'):

    """Returns color parameters for false coloring data.

    Parameters
    ----------

    key : str
        Defaults to 'HE'. Which RGB settings to use, when 'HE' will use
        RGB settings for virtual H&E staining.

        'IHC' will return color settings for virtual IHC/DAB staining.


    Returns
    -------
    color_dict : dict
        Dictionary with keys ('nuclei', 'cyto') or ('nuclei', 'anti'),
        which correspond to lists containing empirically derived
        RGB constants for false coloring.


    """

    color_dict = {
                    'HE': {'nuclei': [0.17, 0.27, 0.105],
                           'cyto': [0.05, 1.0, 0.54]},

                    'IHC': {'nuclei': [0.65, 0.45, 0.15],
                            'anti': [0.4, 0.7, 0.9]}
                }

    return color_dict[key]


def applyCLAHE(image, clahe=None,
               tileGridSize=(8, 8),
               input_dtype=numpy.uint16,
               clipLimit=0.048):
    """
    Applies Contrast Limited Adaptive Histogram Equalization algorithm
    from OpenCV.

    Parameters
    ----------

    image : 2D numpy array
        Image for histogram equalization

    clahe : None or cv2.CLAHE object
        CV2 object to use for equalization

    tileGridSize : tuple
        Tuple of ints representing the windowsize for CLAHE,
        default is (32,32)

    input_dtype : numpy dtype
        Dtype to use for CLAHE object, defaults to numpy.uint16,
        cv2 CLAHE is compatible with either numpy.uint8 or numpy.uint16

    Returns
    -------

    equalized_image : 2D numpy array
        Image with equalized historgram.

    """

    if clahe is None:
        # create clahe object
        clahe = cv2.createCLAHE(tileGridSize=tileGridSize,
                                clipLimit=clipLimit)

    # ensure image is of uint dtype
    image = image.astype(input_dtype)

    # apply CLAHE
    equalized_image = clahe.apply(image)

    # Renormalize to original image levels
    final_image = (image.max())*(equalized_image/equalized_image.max())

    return final_image.astype(input_dtype)


def getBackgroundLevels(image, threshold=50):
    """
    Calculate foreground and background values based on image
    statistics, background is currently set to be 20% of foreground.

    Parameters
    ----------

    image : 2D numpy array

    threshold : int
        threshold above which is counted as foreground

    Returns
    -------

    hi_val : int
        Foreground values

    background : int
        Background value
    """

    image_DS = numpy.sort(image, axis=None)

    foreground_vals = image_DS[numpy.where(image_DS > threshold)]

    hi_val = foreground_vals[int(numpy.round(len(foreground_vals)*0.95))]

    background = hi_val/5

    return hi_val, background


def getFlatField(*args):
    print('depreciated use getIntensityMap instead')


def getIntensityMap(image, tileSize=256, blockSize=16, bgThreshold=50):

    """
    Returns downsampled 3D intensity leveling map of image data by
    breaking image data up into equal sized cubes. Areas which fall
    beneath the background threshold will be set to median the median
    intensity value of the data cube.

    Parameters
    ----------

    image : 2D or 3D numpy array

    tileSize : int
        default is 256. Lateral size for data partition. Smaller
        tileSize will result in higher resolution 3D intensity maps.

    blockSize : int
        default is 16. The final size of the downsampled map will be the
        tileSize divided by blockSize.

    Returns
    -------

    intensityMap : 3D numpy array
        Calculated intensity map for input image

    """

    midrange, background = getBackgroundLevels(image, threshold=bgThreshold)

    rows_max = int(numpy.ceil(image.shape[0]/blockSize)*blockSize)
    cols_max = int(numpy.ceil(image.shape[2]/blockSize)*blockSize)
    stacks_max = int(numpy.ceil(image.shape[1]/blockSize)*blockSize)

    rows = numpy.arange(0, rows_max+int(tileSize/blockSize),
                        int(tileSize/blockSize))

    cols = numpy.arange(0, cols_max+int(tileSize/blockSize),
                        int(tileSize/blockSize))

    stacks = numpy.arange(0, stacks_max+int(tileSize/blockSize),
                          int(tileSize/blockSize))

    intensityMap = numpy.zeros((len(rows)-1, len(stacks)-1,
                                len(cols)-1), dtype=float)

    for i in range(1, len(rows)):
        for j in range(1, len(stacks)):
            for k in range(1, len(cols)):

                ROI_0 = image[rows[i-1]:rows[i],
                              stacks[j-1]:stacks[j],
                              cols[k-1]:cols[k]]

                fkg_ind = numpy.where(ROI_0 > background)
                if fkg_ind[0].size == 0:
                    Mtemp = midrange
                else:
                    Mtemp = numpy.median(ROI_0[fkg_ind])
                intensityMap[i-1, j-1, k-1] = Mtemp + \
                    intensityMap[i-1, j-1, k-1]

    return intensityMap


def interpolateDS(image, k, tileSize=256, beta=1.0):
    """
    Method for resizing downsampled data to be the same size as full
    resolution data. Used for interpolating flat field images.

    Parameters
    ----------

    image : 2D numpy array
        Downsampled data

    k : int
        Index for image location in full res data

    tileSize : int
        Default = 256, block size for interpolation

    beta : float
        Default = 1.0, multiplicative constant for final interpolated
        data.

    Returns
    -------

    C_final : 2D numpy array
        Rescaled downsampled data

    """

    x0 = numpy.floor(k/tileSize)
    x1 = numpy.ceil(k/tileSize)
    x = k/tileSize

    # find region in downsampled data
    if k < int(image.shape[1]*tileSize-tileSize):
        if k < int(tileSize/2):
            C_img = image[:, 0, :]

        elif x0 == x1:
            C_img = image[:, int(x1), :]
        else:
            img_norm0 = image[:, int(x0), :]
            img_norm1 = image[:, int(x1), :]

            # average between two indicies
            C_img = img_norm0 + (x-x0)*(img_norm1 - img_norm0)/(x1-x0)
    else:
        C_img = image[:, image.shape[1]-1, :]

    # interpolate flat fields
    C_final = beta*nd.interpolation.zoom(C_img, tileSize, order=1,
                                         mode='nearest')

    return C_final


def deconvolveColors(image):
    """
    Separates H&E channels from an RGB image using skimage.color.rgb2hed
    method.

    Parameters
    ----------

    image : 3D numpy array
        RGB image in the format [X, Y, C] where the hematoxylin and
        eosin channels are to be separted.

    Returns
    -------

    hematoxylin : 2D numpy array
        nuclear channel deconvolved from RGB image


    eosin : 2D numpy array
        cytoplasm channel deconvolved from RGB image

    """

    separated_image = rgb2hed(image)

    hematoxylin = separated_image[:, :, 0]

    eosin = separated_image[:, :, 1]

    return hematoxylin, eosin


def segmentNuclei(image, return3D=True,
                  opening=True,
                  radius=3,
                  min_size=64,
                  return_cyto=False):
    """

    Grabs binary mask of nuclei from H&E RGB image using color
    deconvolution.

    Parameters
    ----------

    image : 3D numpy array
        H&E stained RGB image in the form [X, Y, C]

    return3D : bool,
        Defaults to False, return 3D version of mask

    return_cyto : bool
        Defaults to False, will return a binary mask for cytoplasm from
        color deconvolved RGB image.

    Returns
    -------

    binary_mask : 2D or 3D numpy array
        Binary mask of segmented nuclei

    """

    # separate channels
    nuclei, cyto = deconvolveColors(image)

    # median filter nuclei for optimized otsu threshold
    median_filtered_nuclei = filt.median(nuclei)

    # calculate threshold and create initial binary mask
    threshold = filt.threshold_otsu(median_filtered_nuclei)
    binarized_nuclei = (median_filtered_nuclei > threshold).astype(int)

    # remove small objects
    labeled_mask = morph.label(binarized_nuclei)
    shape_filtered_mask = morph.remove_small_objects(labeled_mask,
                                                     min_size=min_size)

    # binary opening to separate objects
    if opening:
        shape_filtered_mask = morph.binary_opening(shape_filtered_mask,
                                                   morph.disk(radius))

    # remove labels from nuclear mask
    binary_mask = (shape_filtered_mask > 0)

    # create a binary mask for cytoplasm
    if return_cyto:

        # median filter cyto for optimized otsu threshold
        median_filtered_cyto = filt.median(cyto)

        # calculate threshold and create initial binary mask
        cyto_threshold = filt.threshold_otsu(median_filtered_cyto)

        # create binary mask
        binary_cyto = (median_filtered_cyto > cyto_threshold).astype(int)

        # ensure that nuclei are segmented out of cyto mask
        binary_cyto = binary_cyto*(binary_mask < 1)

        # binary closing to remove pepper noise
        if opening:
            binary_cyto = morph.binary_closing(binary_cyto, morph.disk(radius))

        # create 3D array and rearrange shape to match an RGB image
        if return3D:
            binary_cyto = numpy.moveaxis(numpy.asarray([binary_cyto,
                                                        binary_cyto,
                                                        binary_cyto]), 0, -1)

    # create 3D array and rearrange shape to match an RGB image
    if return3D:
        binary_mask = numpy.moveaxis(numpy.asarray([shape_filtered_mask,
                                                    shape_filtered_mask,
                                                    shape_filtered_mask]),
                                     0, -1)

    # return mask
    if return_cyto:
        return binary_mask.astype(int), binary_cyto.astype(int)

    else:
        return binary_mask.astype(int)


def maskEmpty(image_RGB,
              mask_val=0.05,
              return3D=True,
              min_size=150):

    """
    Method to remove white areas from RGB histology image.

    Parameters
    ----------

    image_RGB : 3D numpy array
        RGB image in the form [X, Y, C].

    mask_val : float
        Value over which pixels will be masked out of hsv image in
        value space.

    return3D : bool
        defaults to True. If True a 3D binary mask will be returned,
        otherwise mask is 2D.

    min_size : int
        Minimum sized object for the area filter. Objects smaller than
        this threshold will be removed.

    Returns
    -------

    empty_mask : numpy array
        Binary mask of empty spaces in an image.
    """

    # convert rgb image to hsv space
    hsv = rgb2hsv(image_RGB)

    # mask white areas
    binary_mask = (hsv[:, :, 1] < mask_val).astype(int)

    # remove small objects and fill holes
    labeled_mask = morph.label(binary_mask)

    labeled_mask = morph.remove_small_objects(labeled_mask,
                                              min_size=min_size)

    labeled_mask = morph.remove_small_holes(labeled_mask)

    empty_mask = (labeled_mask < 1).astype(int)

    # return mask
    if return3D:
        empty_mask_3D = numpy.ones(image_RGB.shape, dtype=int)

        for i in range(empty_mask_3D.shape[-1]):
            empty_mask_3D[:, :, i] *= empty_mask

        return empty_mask_3D

    else:

        return empty_mask



"""
#===============================================================================
#
#  License: GPL
#
#
#  Copyright (c) 2019 Rob Serafin, Liu Lab,
#  The University of Washington Department of Mechanical Engineering
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#===============================================================================

Rob Serafin
11/20/2019

"""

import os
from skimage.io import imread
import numpy
from pathos.multiprocessing import ProcessingPool
import h5py as hp


class DataObject(object):
    def __init__(self, directory, imageSet=None,
                 setupPool=False, ncpus=2, tissue_type='Default'):
        """
        Object to store image data in a convienient way for batch
        processing.

        Attributes
        ----------

        directory : string or pathlike
            Base directory where image data is stored

        imageSet : 3d numpy array
            images for processing

        setupPool : bool
            setup processing pool

        """

        # object base directory
        self.directory = directory

        # object data to process
        self.imageSet = imageSet

        # dedicated cpus
        if setupPool:
            self.setupProcessing(ncpus=ncpus)
        else:
            self.unloadPool()

        # Tissue type for RGB settings
        self.tissue = tissue_type

    def loadImages(self, file_list):
        """
        Loads list of images and returns as a 3D numpy array

        Parameters
        ----------

        file_list : list
            List of filepaths to be read into memory

        Returns
        -------

        images : numpy array
            Image data read into memory
        """

        file_list = sorted(file_list)
        images = []

        for item in file_list:
            images.append(imread(item))

        return numpy.asarray(images)

    def loadH5(self, folder, dataID,
               channelIDs=['s00', 's01'],
               start_index=0, stop_index=0):

        """
        Parameters
        ----------

        folder : str or pathlike
            Folder to grab image data from

        dataID : int
            Resolution of data to grab from HDF5 file. Defaults to zero.

        channelIDs : list
            keys for data entry for HDF5

        start_index : int
            Index to begin reading data from, defaults to zero.

        stop_index : int
            Index to stop reading data from, if zero will continue
            through the entire dataset.

        Returns
        -------

        nuclei : 3D numpy array
            First channel image data from HDF5 file

        cyto : 3D numpy array
            Second channel image data from HDF5 file
        """

        data_name = [os.path.join(folder, f) for f in os.listdir(folder)
                     if f.endswith('h5')]

        if start_index == stop_index:

            with hp.File(data_name[0], 'r') as f:
                nuclei = f['t00000'][channelIDs[0]][str(dataID)]['cells'][:]
                cyto = f['t00000'][channelIDs[1]][str(dataID)]['cells'][:]
            f.close()

        else:

            with hp.File(data_name[0], 'r') as f:
                nuclei = f['t00000'][channelIDs[0]][str(dataID)]['cells'][start_index:stop_index]
                cyto = f['t00000'][channelIDs[1]][str(dataID)]['cells'][start_index:stop_index]
            f.close()

        return nuclei, cyto

    def setupH5data(self, folder=None, dataID=0,
                    channelIDs=['s00', 's01'],
                    start_index=0, stop_index=0):

        """
        Sets up two channel H5 dataset with default key entries

        Parameters
        ----------

        folder : str or pathlike
            Folder to grab image data from, defaults to None. If None
            DataObject will use self.directory.

        dataID : int
            Resolution of data to grab from HDF5 file. Defaults to zero.

        channelIDs : list
            keys for data entry for HDF5

        start_index : int
            Index to begin reading data from, defaults to zero.

        stop_index : int
            Index to stop reading data from, if zero will continue
            through the entire dataset.


        Returns
        -------

        dataset : tuple
            Image data read from HDF5, in the order
            (channel1_data, channel2_data), where each entry is a numpy
            array of image data.

        """
        if folder:
            dataset = self.loadH5(folder, dataID=dataID,
                                  channelIDs=channelIDs)

        else:
            dataset = self.loadH5(self.directory, dataID=dataID,
                                  channelIDs=channelIDs)

        self.imageSet = numpy.asarray(dataset)

    def setupProcessing(self, ncpus):
        """
        Creates processing pool with specified ncpus for DataObject.

        Parameters
        ----------

        ncpus : int
            Number of cpu cores for processing pool

        """

        self.pool = ProcessingPool(ncpus=ncpus)

    def unloadPool(self):
        """
        Turns off processing pool.

        Parameters
        ----------

        Returns
        -------
        """
        self.pool = None

    def processImages(self, runnable_dict, imageSet, dtype=None):
        """
        Method to batch process multiple images simultaneously. Can
        process multiple channels or one at a time. Method acts on
        Image data within data object.

        Parameters
        -------

        runnable_dict : dict
        Dictionary with the following key, value pairs:

            'runnable' : method to run by processingpool.map

            'kwargs' : key word arguments for method

        imageSet : numpy array
            Image data of 3 or more dimmensions, should be in the
            shape [[Z1, X1, Y1], [Z2, X2, Y2], ...etc].

        dtype : None or datatype
            Defaults to None type, if not none data will be returned
            as specified type.

        Returns
        -------

        processed_images : numpy array
            Images which have been procesed using the
            runnable_dict's method.

        """
        if self.pool is None:
            self.setupProcessing(ncpus=4)

        func, kwargs = runnable_dict['runnable'], runnable_dict['kwargs']

        processed_images = []

        if type(kwargs) == dict:
            processed_images.append(self.pool.map(func, *imageSet,
                                                  **kwargs))

        else:
            processed_images.append(self.pool.map(func, *imageSet))

        if dtype is None:
            return numpy.asarray(processed_images)[0]

        else:
            return numpy.asarray(processed_images, dtype=dtype)[0]



"""
#===============================================================================
# 
#  License: GPL
#
#
#  Copyright (c) 2019 Rob Serafin, Liu Lab, 
#  The University of Washington Department of Mechanical Engineering  
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License 2
#  as published by the Free Software Foundation.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#   You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# 
#===============================================================================

Rob Serafin
12/17/2019

"""

import os
import h5py as h5
import numpy
import os
import FalseColor.Color as fc
from FalseColor.SaveThread import saveProcess
import numpy 
from scipy import ndimage
import copy
import argparse
import h5py as h5
import time



class DataQueue(object):
    def __init__(self, directory, filename):
        pass


from setuptools import setup

with open('requirements.txt','r') as f:
    requires = f.read().splitlines()
f.close()

setup(name='falsecolor',
    packages = ['falsecolor'],
    version='1.1.4.7',
    license = 'GNU Affero General Public License v3.0',
    description='Methods for H&E pseudo coloring of grayscale fluorescent images',
    author='Robert Serafin',
    author_email='serrob23@uw.edu',
    url = 'https://github.com/serrob23/FalseColor',
    download_url = 'https://github.com/serrob23/falsecolor/releases/download/v1.1.4.6/falsecolor-1.1.4.6.tar.gz',
    keywords = ['Image Analysis', 
                            'Fluorescence Microscopy', 
                            'Color Translation', 
                            'Virtual Staining',
                            'Histology'],
    install_requires = requires,
    python_requires='>=3.6',

)



"""
#===============================================================================
#
#  License: GPL
#
#
#  Copyright (c) 2019 Rob Serafin, Liu Lab,
#  The University of Washington Department of Mechanical Engineering
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#===============================================================================

Rob Serafin
2/26/2020

"""
import os
from skimage import io


def saveProcess(queue):
    """
    Parameters
    ----------

    queue : multiprocessing queue

        message : result of queue.get(), has the following properties in
        this order:

                path : str or pathlike
                    top level storage directory for data

                folder : str
                    specific dir to save data in

                filename : str
                    "file.tif" filename for data

                data : numpy array
                    image to save

                do_stats : bool, False
                    whether to save RGB statistics, defaults to False

                token : None or str
                    token will be a str when thread stop is called

    Returns
    -------
    """
    while True:

        message = queue.get()

        if message[-1] is not None:
            break

        else:
            (path, folder, filename, data, token) = message

            storage_dir = os.path.join(path, folder)

            if not os.path.exists(storage_dir):
                os.mkdir(storage_dir)

            file_savename = os.path.join(storage_dir, filename)
            print(storage_dir, filename)

            io.imsave(file_savename, data)

            message = None
